# gmab/providers/__init__.py

from gmab.providers.linode import LinodeProvider
from gmab.providers.aws import AWSProvider
from gmab.providers.hetzner import HetznerProvider

def get_provider(provider_name, provider_cfg):
    """
    Factory function to get the appropriate provider instance.
    
    Args:
        provider_name (str): The name of the provider ('linode', 'aws', or 'hetzner')
        provider_cfg (dict): Provider configuration
        
    Returns:
        ProviderBase: A provider instance
        
    Raises:
        ValueError: If the provider name is unknown or if provider config is invalid
    """
    providers = {
        "linode": LinodeProvider,
        "aws": AWSProvider,
        "hetzner": HetznerProvider
    }
    
    provider_class = providers.get(provider_name)
    if not provider_class:
        raise ValueError(f"Unknown provider: {provider_name}")
    
    if not provider_cfg:
        raise ValueError(f"Missing or empty configuration for provider: {provider_name}")
    
    provider = provider_class(provider_cfg)
    provider.provider_name = provider_name  # Set the provider name
    return provider


# gmab/providers/hetzner.py

import click
import functools
import requests
import random
import string
import time
from pathlib import Path
from gmab.providers.base import ProviderBase

def generate_random_string(length=12):
    """Generate a random string of lowercase letters and digits."""
    return ''.join(random.choices(string.ascii_lowercase + string.digits, k=length))

class HetznerProvider(ProviderBase):
    """
    Provider implementation for Hetzner Cloud.
    """
    
    def __init__(self, provider_cfg):
        """
        Initialize the Hetzner provider with the given configuration.
        
        Args:
            provider_cfg (dict): Hetzner provider configuration including API token, etc.
        """
        super().__init__(provider_cfg)
        
        # Validate required config values
        api_key = provider_cfg.get('api_key')
        if not api_key:
            raise ValueError("Hetzner requires an API key to be configured")
            
        self.api_url = "https://api.hetzner.cloud/v1"
        self.headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json"
        }

    def _get_instance_expiry_info(self, labels):
        """
        Helper method to get expiry information from instance labels.
        
        Args:
            labels (dict): Key-value labels from the Hetzner instance
            
        Returns:
            tuple: (creation_time, lifetime_minutes, is_expired)
        """
        creation_time = 0
        lifetime_minutes = 60
        
        # In Hetzner, we store these as gmab-creation-time and gmab-lifetime
        creation_time = int(labels.get("gmab-creation-time", "0"))
        lifetime_minutes = int(labels.get("gmab-lifetime", "60"))
        
        current_time = int(time.time())
        is_expired = (current_time - creation_time) > (lifetime_minutes * 60)
        return creation_time, lifetime_minutes, is_expired

    def _get_or_create_ssh_key(self, ssh_key_content):
        """
        Helper method to get existing SSH key or create a new one.
        
        Args:
            ssh_key_content (str): The SSH public key content
            
        Returns:
            int: SSH key ID
            
        Raises:
            Exception: If SSH key cannot be created or found
        """
        try:
            # First list existing SSH keys
            list_response = requests.get(
                f"{self.api_url}/ssh_keys",
                headers=self.headers,
                timeout=30
            )
            
            if list_response.status_code != 200:
                raise Exception(f"Failed to list SSH keys: {list_response.text}")
                
            # Check if we have a matching key
            ssh_keys = list_response.json()["ssh_keys"]
            for key in ssh_keys:
                if key["public_key"].strip() == ssh_key_content.strip():
                    return key["id"]
            
            # If no matching key found, create new one
            ssh_key_name = f"gmab-key-{generate_random_string(8)}"
            create_response = requests.post(
                f"{self.api_url}/ssh_keys",
                headers=self.headers,
                json={
                    "name": ssh_key_name,
                    "public_key": ssh_key_content
                },
                timeout=30
            )

            if create_response.status_code != 201:
                raise Exception(f"Failed to create SSH key: {create_response.text}")

            return create_response.json()["ssh_key"]["id"]
            
        except requests.exceptions.RequestException as e:
            raise Exception(f"Network error when managing SSH keys: {str(e)}")
        except Exception as e:
            raise Exception(f"Failed to get or create SSH key: {str(e)}")
        
    @staticmethod
    def get_default_config():
        return {
            "api_key": "",
            "default_region": "nbg1",
            "default_image": "ubuntu-22.04",
            "default_type": "cpx11"
        }
    
    @staticmethod
    def get_config_prompts(provider_config):
        config = {}
        config['api_key'] = functools.partial(click.prompt,
            "API Key",
            default=provider_config.get('api_key', ''),
            hide_input=False
        )
        config['default_region'] = functools.partial(click.prompt,
            "Default region",
            default=provider_config.get('default_region', HetznerProvider.get_default_config()['default_region'])
        )
        config['default_image'] = functools.partial(click.prompt,
            "Default image",
            default=provider_config.get('default_image', HetznerProvider.get_default_config()['default_image'])
        )
        config['default_type'] = functools.partial(click.prompt,
            "Default instance type",
            default=provider_config.get('default_type', HetznerProvider.get_default_config()['default_type'])
        )
        return config

    def spawn_instance(self, image=None, region=None, ssh_key_path=None, lifetime_minutes=None):
        """
        Create a new Hetzner Cloud server.
        
        Args:
            image (str, optional): The Hetzner image to use
            region (str, optional): The location to create the server in
            ssh_key_path (str, optional): Path to the SSH public key to use
            lifetime_minutes (int, optional): Lifetime of the server in minutes
            
        Returns:
            dict: Server information including ID, IP, etc.
            
        Raises:
            ValueError: If API key is missing
            FileNotFoundError: If SSH key file doesn't exist
            Exception: For API errors or other failures
        """
        # Use provided values or fall back to defaults
        default_type = self.provider_cfg.get("default_type", "cpx11")
        default_image = self.provider_cfg.get("default_image", "ubuntu-22.04")
        default_region = self.provider_cfg.get("default_region", "nbg1")

        chosen_image = image or default_image
        chosen_region = region or default_region
        
        if lifetime_minutes is None:
            lifetime_minutes = 60

        # Current timestamp for creation time
        creation_time = int(time.time())

        # Read SSH key
        ssh_key_path = ssh_key_path or self.provider_cfg.get("ssh_key_path", "~/.ssh/id_ed25519.pub")
        keyfile = Path(ssh_key_path).expanduser()
        if not keyfile.exists():
            raise FileNotFoundError(f"SSH key not found at {keyfile}")

        with open(keyfile, 'r') as f:
            ssh_key_content = f.read().strip()

        # Generate a unique name
        instance_name = f"gmab-{generate_random_string(12)}"

        try:
            # Get or create SSH key
            ssh_key_id = self._get_or_create_ssh_key(ssh_key_content)

            # Create the server with Hetzner-compliant labels
            create_response = requests.post(
                f"{self.api_url}/servers",
                headers=self.headers,
                json={
                    "name": instance_name,
                    "server_type": default_type,
                    "image": chosen_image,
                    "location": chosen_region,
                    "ssh_keys": [ssh_key_id],
                    "labels": {
                        "gmab": "true",
                        "gmab-creation-time": str(creation_time),
                        "gmab-lifetime": str(lifetime_minutes)
                    }
                },
                timeout=60  # Server creation can take a bit longer
            )

            if create_response.status_code != 201:
                raise Exception(f"Failed to create server: {create_response.text}")

            server_data = create_response.json()["server"]
            
            return {
                "provider": self.provider_name,
                "instance_id": str(server_data["id"]),
                "label": instance_name,
                "ip": server_data.get("public_net", {}).get("ipv4", {}).get("ip", "No IP Assigned"),
                "status": server_data["status"],
                "region": chosen_region,
                "image": chosen_image,
                "creation_time": creation_time,
                "lifetime_minutes": lifetime_minutes
            }
            
        except requests.exceptions.RequestException as e:
            raise Exception(f"Network error when creating server: {str(e)}")
        except Exception as e:
            raise Exception(f"Failed to create Hetzner instance: {str(e)}")

    def get_instance_id_by_label(self, label):
        """
        Find instance ID by label, but only for instances with the 'gmab' label.
        
        Args:
            label (str): The instance label to search for
            
        Returns:
            str or None: The instance ID if found, None otherwise
        """
        try:
            instances = self.list_instances()
            for instance in instances:
                if instance["label"] == label:
                    return instance["instance_id"]
            return None
            
        except Exception:
            return None

    def terminate_instance(self, instance_identifier):
        """
        Terminate a Hetzner server by ID or label.
        
        Args:
            instance_identifier (str): Either a numeric instance ID or a label
            
        Raises:
            Exception: If the server cannot be found or deleted
        """
        try:
            # If it's not a numeric ID, try to find by label
            if not instance_identifier.isdigit():
                instance_id = self.get_instance_id_by_label(instance_identifier)
                if instance_id is None:
                    raise Exception(f"No instance found with label '{instance_identifier}'")
            else:
                instance_id = instance_identifier

            # Delete the server
            delete_response = requests.delete(
                f"{self.api_url}/servers/{instance_id}",
                headers=self.headers,
                timeout=30
            )

            if delete_response.status_code not in (200, 204):
                raise Exception(f"Failed to delete server: {delete_response.text}")
                
        except requests.exceptions.RequestException as e:
            raise Exception(f"Network error when terminating server: {str(e)}")
        except Exception as e:
            raise Exception(f"Failed to terminate Hetzner instance: {str(e)}")

    def list_instances(self):
        """
        List all Hetzner servers tagged with 'gmab'.
        
        Returns:
            list: List of instance dictionaries
            
        Raises:
            Exception: If listing servers fails
        """
        try:
            response = requests.get(
                f"{self.api_url}/servers",
                headers=self.headers,
                params={"label_selector": "gmab"},
                timeout=30
            )

            if response.status_code != 200:
                raise Exception(f"Failed to list servers: {response.text}")

            instances = []
            for server in response.json()["servers"]:
                creation_time, lifetime_minutes, is_expired = self._get_instance_expiry_info(server.get("labels", {}))
                
                # Modify status to include expiry information
                base_status = server["status"]
                status = f"{base_status} (expired)" if is_expired else base_status

                instances.append({
                    "provider": self.provider_name,
                    "instance_id": str(server["id"]),
                    "label": server["name"],
                    "ip": server.get("public_net", {}).get("ipv4", {}).get("ip", "No IP Assigned"),
                    "status": status,
                    "region": server["datacenter"]["location"]["name"],
                    "image": server["image"]["name"],
                    "creation_time": creation_time,
                    "lifetime_minutes": lifetime_minutes,
                    "is_expired": is_expired
                })

            return instances
            
        except requests.exceptions.RequestException as e:
            raise Exception(f"Network error when listing servers: {str(e)}")
        except Exception as e:
            raise Exception(f"Failed to list Hetzner instances: {str(e)}")

    def list_expired_instances(self):
        """
        List all expired instances.
        
        Returns:
            list: List of expired instance dictionaries
        """
        return [inst for inst in self.list_instances() if inst['is_expired']]


# gmab/cli.py

import click
from gmab.commands.spawn import spawn_box
from gmab.commands.terminate import terminate_box
from gmab.commands.list import list_boxes
from gmab.commands.configure import run_configure, print_configs
from gmab.utils.config_loader import DEFAULT_PROVIDERS_CONFIG, config_exists, ConfigNotFoundError, load_config
from gmab import __version__

def check_config_exists():
    """Check if config exists and show an error message if it doesn't."""
    if not config_exists():
        click.echo("Error: GMAB is not configured.")
        click.echo("Please run 'gmab configure' to set up your configuration.")
        return False
    return True

def get_configured_providers():
    """Get a list of providers that have been explicitly configured."""
    try:
        providers_config = load_config("providers.json")
        return list(providers_config.keys())
    except (ConfigNotFoundError, Exception):
        return []

@click.group(invoke_without_command=True)
@click.version_option(version=__version__, prog_name='gmab',message='%(prog)s %(version)s')
@click.pass_context
def cli(ctx):
    """ gmab (Give Me A Box) - CLI tool to spawn, list, and manage temporary cloud boxes. """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())

@cli.command()
@click.option('--provider', '-p', default=None, help='Provider name (default from config).')
@click.option('--region', '-r', default=None, help='Override region (default from config).')
@click.option('--image', '-i', default=None, help='Override image (default from config).')
@click.option('--lifetime', '-t', type=int, default=None, help='Lifetime in minutes (default: 60).')
def spawn(provider, region, image, lifetime):
    """ Spawn a new instance. """
    if not check_config_exists():
        return
    try:
        if provider and provider not in get_configured_providers():
            click.echo(f"Error: Provider '{provider}' is not configured.")
            click.echo(f"Please run 'gmab configure -p {provider}' to configure this provider.")
            return
            
        spawn_box(provider, region, image, lifetime)
    except ConfigNotFoundError:
        click.echo("Error: GMAB is not configured.")
        click.echo("Please run 'gmab configure' to set up your configuration.")
    except Exception as e:
        click.echo(f"Error: {str(e)}")

@cli.command()
@click.argument('instance_ids', nargs=-1)
@click.option('--provider', '-p', default=None, help='Provider name (default from config).')
@click.option('--yes', '-y', is_flag=True, help='Skip confirmation prompt.')
def terminate(instance_ids, provider, yes):
    """ Terminate one or more instances by ID or label. 
    
    Use 'all' to terminate all instances, or 'expired' to terminate expired instances.
    """
    if not check_config_exists():
        return
    
    try:
        if provider and provider not in get_configured_providers():
            click.echo(f"Error: Provider '{provider}' is not configured.")
            click.echo(f"Please run 'gmab configure -p {provider}' to configure this provider.")
            return
            
        # Handle 'expired' command
        if len(instance_ids) == 1 and instance_ids[0] == 'expired':
            # Get all instances
            instances = list_boxes(provider)
            if not instances:
                click.echo("No active instances found.")
                return
            
            # Filter expired instances
            expired_instances = [inst for inst in instances if inst.get('is_expired', False)]
            
            if not expired_instances:
                click.echo("No expired instances found.")
                return
            
            # Show what will be terminated
            click.echo("The following expired instances will be terminated:")
            for instance in expired_instances:
                click.echo(f"- {instance['instance_id']} ({instance['provider']}: {instance['label']})")
            
            # Ask for confirmation unless -y flag is used
            if not yes and not click.confirm("Do you want to proceed?"):
                click.echo("Operation cancelled.")
                return
            
            # Terminate expired instances
            success_count = 0
            failed_instances = []
            for instance in expired_instances:
                try:
                    terminate_box(instance['instance_id'], instance['provider'])
                    success_count += 1
                except Exception as e:
                    failed_instances.append((instance['instance_id'], str(e)))
            
            # Print summary
            if success_count > 0:
                click.echo(f"Successfully terminated {success_count} expired instance(s).")
            if failed_instances:
                click.echo("\nFailed to terminate the following instances:")
                for instance_id, error in failed_instances:
                    click.echo(f"- {instance_id}: {error}")
            
            return

        # Handle 'all' command
        if len(instance_ids) == 1 and instance_ids[0] == 'all':
            # Get all instances
            instances = list_boxes(provider)
            if not instances:
                click.echo("No active instances found.")
                return
            
            # Show what will be terminated
            click.echo("The following instances will be terminated:")
            for instance in instances:
                click.echo(f"- {instance['instance_id']} ({instance['provider']}: {instance['label']})")
            
            # Ask for confirmation unless -y flag is used
            if not yes and not click.confirm("Do you want to proceed?"):
                click.echo("Operation cancelled.")
                return
            
            # Terminate all instances
            success_count = 0
            failed_instances = []
            for instance in instances:
                try:
                    terminate_box(instance['instance_id'], instance['provider'])
                    success_count += 1
                except Exception as e:
                    failed_instances.append((instance['instance_id'], str(e)))
            
            # Print summary
            if success_count > 0:
                click.echo(f"Successfully terminated {success_count} instance(s).")
            if failed_instances:
                click.echo("\nFailed to terminate the following instances:")
                for instance_id, error in failed_instances:
                    click.echo(f"- {instance_id}: {error}")
            
            return
        
        # Regular termination logic
        if not instance_ids:
            click.echo("No instance IDs or labels provided.")
            return
            
        if len(instance_ids) > 5:
            click.echo("Error: Cannot terminate more than 5 instances at once.")
            return

        if len(instance_ids) > 1 and not yes:
            # Ask for confirmation when terminating multiple instances
            instances_str = "', '".join(instance_ids)
            if not click.confirm(f"Are you sure you want to terminate instances: '{instances_str}'?"):
                click.echo("Operation cancelled.")
                return

        success_count = 0
        failed_instances = []
        for instance_id in instance_ids:
            try:
                terminate_box(instance_id, provider)
                success_count += 1
            except Exception as e:
                failed_instances.append((instance_id, str(e)))

        # Print summary
        if success_count > 0:
            click.echo(f"Successfully terminated {success_count} instance(s).")
        if failed_instances:
            click.echo("\nFailed to terminate the following instances:")
            for instance_id, error in failed_instances:
                click.echo(f"- {instance_id}: {error}")

    except ConfigNotFoundError:
        click.echo("Error: GMAB is not configured.")
        click.echo("Please run 'gmab configure' to set up your configuration.")
    except Exception as e:
        click.echo(f"Error: {str(e)}")

@cli.command(name='list')  # Explicitly name the command
@click.option('--provider', '-p', 
              help='Provider name (list all providers if not specified).')
def list_command(provider):  # Renamed from 'list' to avoid conflicts
    """ List active instances. """
    if not check_config_exists():
        return
    
    try:
        # Check if provider is configured
        if provider and provider not in get_configured_providers():
            click.echo(f"Error: Provider '{provider}' is not configured.")
            click.echo(f"Please run 'gmab configure -p {provider}' to configure this provider.")
            return
            
        instances = list_boxes(provider)

        if not instances or len(instances) == 0:
            click.echo("No active instances found.")
            return

        # Updated column widths
        columns = {
            'Provider': 10,
            'Instance ID': 22,
            'Label': 20,
            'IP Address': 18,
            'Status': 20,  # Increased to accommodate "(expired)" suffix
            'Region': 15,
            'Image': 25,
            'Time Left': 15  # Changed from 'Lifetime' to 'Time Left'
        }

        # Print header
        header = (
            f"{'Provider':<{columns['Provider']}} "
            f"{'Instance ID':<{columns['Instance ID']}} "
            f"{'Label':<{columns['Label']}} "
            f"{'IP Address':<{columns['IP Address']}} "
            f"{'Status':<{columns['Status']}} "
            f"{'Region':<{columns['Region']}} "
            f"{'Image':<{columns['Image']}} "
            f"{'Time Left':<{columns['Time Left']}}"
        )
        click.echo(header)
        click.echo("=" * (sum(columns.values()) + len(columns) - 1))

        for instance in instances:
            # Add safety checks for required fields
            provider = instance.get('provider', 'Unknown')
            instance_id = instance.get('instance_id', 'Unknown')
            label = instance.get('label', 'Unknown')
            ip = instance.get('ip', 'No IP')
            status = instance.get('status', 'Unknown')
            region = instance.get('region', 'Unknown')
            image = instance.get('image', 'Unknown')
            
            # Format lifetime left
            lifetime_left = instance.get('lifetime_left', 0)
            if lifetime_left < 1:
                time_left = "expired"
            else:
                time_left = f"{int(lifetime_left)}m"

            click.echo(
                f"{provider:<{columns['Provider']}} "
                f"{instance_id:<{columns['Instance ID']}} "
                f"{label:<{columns['Label']}} "
                f"{ip:<{columns['IP Address']}} "
                f"{status:<{columns['Status']}} "
                f"{region:<{columns['Region']}} "
                f"{image:<{columns['Image']}} "
                f"{time_left:<{columns['Time Left']}}"
            )

    except ConfigNotFoundError:
        click.echo("Error: GMAB is not configured.")
        click.echo("Please run 'gmab configure' to set up your configuration.")
    except Exception as e:
        click.echo(f"Error: {str(e)}")

@cli.command()
@click.option(
    '--provider', '-p',
    type=click.Choice(['all'] + list(DEFAULT_PROVIDERS_CONFIG.keys())),
    default='all',
    help='Specific provider to configure (default: all providers).'
)
@click.option(
    '--print', 'print_config', is_flag=True,
    help='Print current configuration files and their locations.'
)
def configure(provider, print_config):
    """ Configure GMAB settings and provider credentials. 
    
    This command helps you set up or update your GMAB configuration, including SSH keys,
    provider credentials, and default settings. You can configure all providers or specify
    a single provider to configure.

    The configuration will be stored in your user config directory:
    - Linux/macOS: ~/.config/gmab/
    - Windows: %APPDATA%\\gmab\\
    
    You can override the config location by setting the GMAB_CONFIG_DIR environment variable.
    """
    if print_config:
        print_configs()
    else:
        run_configure(provider)

def main():
    cli()

if __name__ == '__main__':
    main()


# gmab/commands/terminate.py

import click
from gmab.utils.config_loader import load_config, ConfigNotFoundError
from gmab.providers import get_provider

def get_instance_provider(instance_identifier, providers_cfg):
    """
    Determine which provider an instance belongs to based on ID format and provider queries.
    Returns tuple of (provider_name, provider_instance) or (None, None) if not found.
    
    Args:
        instance_identifier (str): The instance ID or label
        providers_cfg (dict): The provider configurations
        
    Returns:
        tuple: (provider_name, provider_instance) or (None, None) if not found
    """
    # First try to determine by ID format
    if instance_identifier.startswith('i-'):
        # AWS instance ID format
        provider_name = "aws"
        if provider_name in providers_cfg:
            provider = get_provider(provider_name, providers_cfg[provider_name])
            return provider_name, provider
    elif instance_identifier.startswith('gmab-'):
        # Check all providers for this label since multiple could use this format
        provider_name = None
    else:
        # For numeric IDs or other formats, we need to check all providers
        provider_name = None

    # If we couldn't determine by format or need to check labels, query each provider
    for provider_name, provider_cfg in providers_cfg.items():
        if not provider_cfg:  # Skip empty provider configs
            continue
            
        provider = get_provider(provider_name, provider_cfg)
        try:
            instances = provider.list_instances()
            
            for instance in instances:
                if (instance['instance_id'] == instance_identifier or 
                    instance['label'] == instance_identifier):
                    return provider_name, provider
        except Exception:
            # Skip providers that fail to list instances
            continue

    return None, None

def terminate_box(instance_id, provider_name=None):
    """
    Terminate an instance by ID or label.
    
    Args:
        instance_id (str): The instance ID or label to terminate
        provider_name (str, optional): The provider name. If None, it will be determined from the instance ID.
        
    Raises:
        ConfigNotFoundError: If configuration files don't exist
        ValueError: If the provider can't be determined or found
        Exception: For other errors
    """
    try:
        # Load configs
        providers_cfg = load_config("providers.json")
        general_cfg = load_config("config.json")

        if provider_name:
            # If provider is specified, use it directly
            provider_cfg = providers_cfg.get(provider_name)
            if not provider_cfg:
                raise ValueError(f"Provider '{provider_name}' is not configured. Run 'gmab configure -p {provider_name}' first.")
            provider = get_provider(provider_name, provider_cfg)
        else:
            # Try to determine provider from instance ID
            provider_name, provider = get_instance_provider(instance_id, providers_cfg)
            if not provider:
                raise ValueError(f"Could not determine provider for instance '{instance_id}'. Make sure the provider is configured.")

        # Terminate the instance
        provider.terminate_instance(instance_id)
        click.echo(f"Terminated instance '{instance_id}' on '{provider_name}'.")
        
    except ConfigNotFoundError:
        # Let the CLI handle this error
        raise
    except Exception as e:
        # Re-raise with a more informative message
        raise Exception(f"Failed to terminate instance: {str(e)}")


# gmab/commands/list.py

from gmab.providers import get_provider
from gmab.utils.config_loader import load_config, ConfigNotFoundError
import time
import click

def get_lifetime_left(instance):
    """Calculate the lifetime left in minutes for an instance."""
    creation_time = instance.get('creation_time', 0)
    lifetime_minutes = instance.get('lifetime_minutes', 60)
    elapsed_minutes = (time.time() - creation_time) / 60
    return max(0, lifetime_minutes - elapsed_minutes)

def list_boxes(provider_name=None):
    """
    Retrieve all gmab-tagged instances from the specified provider(s).
    If no provider is specified, list instances from all configured providers.
    
    Args:
        provider_name (str, optional): The provider to list instances from. If None, list from all providers.
        
    Returns:
        list: A list of instance dictionaries with provider, instance_id, label, etc.
        
    Raises:
        ConfigNotFoundError: If configuration files don't exist
        ValueError: If the specified provider configuration is not found
        Exception: For other errors
    """
    try:
        # Load provider configuration from config file
        providers_config = load_config("providers.json")
        general_config = load_config("config.json")

        instances = []
        
        # If no provider specified, try to list from all configured providers
        if provider_name is None:
            # Now we only iterate over providers that actually have a configuration
            for provider_name, provider_cfg in providers_config.items():
                if not provider_cfg:  # Skip empty configs
                    continue
                    
                try:
                    provider = get_provider(provider_name, provider_cfg)
                    provider_instances = provider.list_instances()
                    instances.extend(provider_instances)
                except Exception as e:
                    click.echo(f"Warning: Failed to list instances from provider '{provider_name}': {str(e)}")
        else:
            # List instances from specific provider
            provider_cfg = providers_config.get(provider_name)
            if not provider_cfg:
                raise ValueError(f"Provider '{provider_name}' is not configured. Run 'gmab configure -p {provider_name}' first.")

            provider = get_provider(provider_name, provider_cfg)
            instances = provider.list_instances()

        # Calculate lifetime left for each instance and add it to the instance data
        for instance in instances:
            instance['lifetime_left'] = get_lifetime_left(instance)

        # Sort instances by lifetime left (descending)
        instances.sort(key=lambda x: x['lifetime_left'], reverse=True)

        return instances
        
    except ConfigNotFoundError:
        # Let the CLI handle this error
        raise
    except Exception as e:
        # Re-raise with a more informative message
        raise Exception(f"Failed to list instances: {str(e)}")


# gmab/utils/config_loader.py

import json
import os
from pathlib import Path
from gmab.utils.paths import get_config_file_path, ensure_config_dir_exists
from gmab.providers import *

# Default configurations
DEFAULT_GENERAL_CONFIG = {
    "ssh_key_path": "~/.ssh/id_ed25519.pub",
    "default_lifetime_minutes": 60,
    "default_provider": "linode"
}

# TODO: Any usage of DEFAULT_PROVIDERS_CONFIG should be replaced by a function
DEFAULT_PROVIDERS_CONFIG = {
    "linode": LinodeProvider.get_default_config(),
    "aws": AWSProvider.get_default_config(),
    "hetzner": AWSProvider.get_default_config(),
}

class ConfigNotFoundError(Exception):
    """Exception raised when a config file does not exist and should not be auto-created."""
    pass

def load_config(filename, create_if_missing=False):
    """
    Load configuration from the appropriate config directory.
    
    Args:
        filename (str): The configuration file to load
        create_if_missing (bool): Whether to create a default config if none exists
    
    Returns:
        dict: The loaded configuration
        
    Raises:
        ConfigNotFoundError: If the config doesn't exist and create_if_missing=False
    """
    # Handle both old-style paths and new config names
    if filename.startswith('gmab/config/'):
        filename = filename.split('/')[-1]

    # Map old filenames to new ones
    filename_map = {
        'general.json': 'config.json',
        'providers.json': 'providers.json'
    }
    
    actual_filename = filename_map.get(filename, filename)
    config_path = get_config_file_path(actual_filename)

    # Check if config exists
    if not config_path.exists():
        # If it doesn't exist and we shouldn't create it, raise an error
        if not create_if_missing:
            raise ConfigNotFoundError(
                f"Configuration file {actual_filename} not found. "
                "Please run 'gmab configure' to set up your configuration."
            )
        
        # Otherwise create default config
        default_content = (DEFAULT_GENERAL_CONFIG if actual_filename == 'config.json' 
                         else DEFAULT_PROVIDERS_CONFIG)
        ensure_config_dir_exists()
        with open(config_path, 'w', encoding='utf-8') as f:
            json.dump(default_content, f, indent=2)

    try:
        with open(config_path, 'r', encoding='utf-8-sig') as f:
            return json.load(f)
    except json.JSONDecodeError as e:
        raise ValueError(f"Error parsing config file {config_path}: {str(e)}")
    except Exception as e:
        raise Exception(f"Error loading config from {config_path}: {str(e)}")

def config_exists():
    """Check if the basic configuration files exist.
    
    Returns:
        bool: True if both config.json and providers.json exist
    """
    config_path = get_config_file_path('config.json')
    providers_path = get_config_file_path('providers.json')
    return config_path.exists() and providers_path.exists()

def save_config(config, filename):
    """Save configuration to file."""
    config_path = get_config_file_path(filename)
    ensure_config_dir_exists()
    with open(config_path, 'w', encoding='utf-8') as f:
        json.dump(config, f, indent=2)


# gmab/commands/configure.py

import click
from pathlib import Path
import json
from gmab.utils.paths import get_config_file_path, ensure_config_dir_exists
from gmab.utils.config_loader import (
    load_config, save_config, DEFAULT_GENERAL_CONFIG, 
    DEFAULT_PROVIDERS_CONFIG, ConfigNotFoundError
)
from gmab import providers

def update_nested_dict(original, updates):
    """Recursively update a nested dictionary without overwriting unspecified values."""
    for key, value in updates.items():
        if key in original and isinstance(original[key], dict) and isinstance(value, dict):
            update_nested_dict(original[key], value)
        elif value is not None:  # Only update if value is not None
            original[key] = value
    return original

def print_configs():
    """Print the current configuration files and their contents."""
    config_files = {
        'config.json': 'General Configuration',
        'providers.json': 'Provider Configuration'
    }

    for filename, description in config_files.items():
        config_path = get_config_file_path(filename)
        click.echo(f"\n{description}")
        click.echo(f"Location: {config_path}")
        
        if config_path.exists():
            try:
                with open(config_path, 'r', encoding='utf-8') as f:
                    config_data = json.load(f)
                    # Handle sensitive data
                    if filename == 'providers.json':
                        # Mask sensitive values in a copy of the config
                        masked_config = json.loads(json.dumps(config_data))
                        for provider in masked_config.values():
                            for key in provider:
                                if key in ['api_key', 'access_key', 'secret_key', 'default_root_pass']:
                                    provider[key] = '********'
                        config_data = masked_config
                    
                    click.echo("Contents:")
                    click.echo(json.dumps(config_data, indent=2))
            except Exception as e:
                click.echo(f"Error reading configuration: {str(e)}")
        else:
            click.echo("File does not exist")

def configure_general(current_config):
    """Configure general settings."""
    click.echo("\nConfiguring general settings:")
    
    ssh_key_path = click.prompt(
        "SSH public key path",
        default=current_config.get('ssh_key_path', DEFAULT_GENERAL_CONFIG['ssh_key_path'])
    )
    
    default_lifetime = click.prompt(
        "Default instance lifetime (minutes)",
        default=current_config.get('default_lifetime_minutes', DEFAULT_GENERAL_CONFIG['default_lifetime_minutes']),
        type=int
    )
    
    # Only show providers that have been configured
    available_providers = list(DEFAULT_PROVIDERS_CONFIG.keys())
    
    # In case we're updating an existing config, get the current provider
    current_provider = current_config.get('default_provider', DEFAULT_GENERAL_CONFIG['default_provider'])
    
    default_provider = click.prompt(
        "Default provider",
        default=current_provider,
        type=click.Choice(available_providers, case_sensitive=False)
    )
    
    return {
        'ssh_key_path': ssh_key_path,
        'default_lifetime_minutes': default_lifetime,
        'default_provider': default_provider
    }

def configure_provider(provider_name, current_config):
    """Configure a specific provider."""
    click.echo(f"\nConfiguring {provider_name} provider:")
    
    provider_config = current_config.get(provider_name, {})
    config = None
    
    # Iterate each class in gmab/providers to get their prompts
    for name, cls in providers.__dict__.items():
        if name.lower() == provider_name.lower() + "provider":
            config = cls.get_config_prompts(provider_config)
    
    # Execute the prompts
    for key in config.keys():
        config[key] = config[key]()
    
    return config
        

def validate_configs():
    """Perform basic validation of the configuration files."""
    try:
        general_config = load_config('config.json')
        providers_config = load_config('providers.json')
        
        # Check if SSH key exists
        ssh_key_path = Path(general_config.get('ssh_key_path', '')).expanduser()
        if not ssh_key_path.exists():
            click.echo(f"\nWarning: SSH key not found at {ssh_key_path}")
        
        # Check if default provider is configured
        default_provider = general_config.get('default_provider')
        if default_provider:
            if default_provider not in providers_config:
                click.echo(f"\nWarning: Default provider '{default_provider}' is not configured")
            else:
                # TODO: This should be done in a validate() function inside the provider
                provider_config = providers_config.get(default_provider, {})
                if default_provider == 'aws':
                    if not provider_config.get('access_key') or not provider_config.get('secret_key'):
                        click.echo(f"\nWarning: Default provider '{default_provider}' is not fully configured (missing AWS credentials)")
                elif not provider_config.get('api_key'):
                    click.echo(f"\nWarning: Default provider '{default_provider}' is not fully configured (missing API key)")
    except Exception as e:
        click.echo(f"\nWarning: Could not validate configurations: {str(e)}")

def run_configure(provider):
    """Main configure function to be called from CLI."""
    # Ensure config directory exists
    config_dir = ensure_config_dir_exists()
    click.echo(f"Using config directory: {config_dir}")
    
    # Load existing configs or create new empty ones
    try:
        general_config = load_config('config.json')
    except ConfigNotFoundError:
        general_config = {}
    
    try:
        providers_config = load_config('providers.json')
    except ConfigNotFoundError:
        providers_config = {}
    
    if provider == 'all':
        # Configure general settings
        new_general_config = configure_general(general_config)
        save_config(new_general_config, 'config.json')
        
        # Configure each provider
        for prov in DEFAULT_PROVIDERS_CONFIG.keys():
            if click.confirm(f"\nDo you want to configure {prov}?", default=True):
                providers_config[prov] = configure_provider(prov, providers_config)
        
        save_config(providers_config, 'providers.json')
    
    else:
        # If no general config exists, configure it first
        if not general_config:
            new_general_config = configure_general({})
            save_config(new_general_config, 'config.json')
        
        # Configure only the specified provider
        providers_config[provider] = configure_provider(provider, providers_config)
        save_config(providers_config, 'providers.json')
    
    # Validate the final configuration
    validate_configs()
    
    click.echo("\nConfiguration completed successfully!")
    click.echo(f"Config files are located in: {config_dir}")


# gmab/commands/spawn.py

import click
from gmab.utils.config_loader import load_config, ConfigNotFoundError
from gmab.providers import get_provider

def spawn_box(provider_name=None, region=None, image=None, lifetime=None):
    """Spawn a new cloud instance with the specified parameters."""
    try:
        # Load main configs
        general_cfg = load_config("config.json")
        providers_cfg = load_config("providers.json")

        # Determine provider
        if not provider_name:
            provider_name = general_cfg.get("default_provider", "linode")

        # Retrieve provider's config
        provider_cfg = providers_cfg.get(provider_name)
        if not provider_cfg:
            raise ValueError(f"Provider '{provider_name}' is not configured. Run 'gmab configure -p {provider_name}' first.")

        # Determine region and image
        chosen_region = region or provider_cfg.get("default_region", "us-east")
        chosen_image = image or provider_cfg.get("default_image", "linode/ubuntu22.04")
        chosen_lifetime = lifetime or general_cfg.get("default_lifetime_minutes", 60)

        # Instantiate provider
        provider = get_provider(provider_name, provider_cfg)

        # Spawn instance
        instance_info = provider.spawn_instance(
            image=chosen_image,
            region=chosen_region,
            ssh_key_path=general_cfg["ssh_key_path"],
            lifetime_minutes=chosen_lifetime
        )

        click.echo(f"Spawned '{provider_name}' instance:")
        click.echo(f"  ID: {instance_info['instance_id']}")
        click.echo(f"  Label: {instance_info['label']}")
        click.echo(f"  IP: {instance_info['ip']}")
        click.echo(f"  Connect via: ssh root@{instance_info['ip']}")
        
        return instance_info
    
    except ConfigNotFoundError:
        # Let the CLI handle this error
        raise
    except Exception as e:
        raise Exception(f"Failed to spawn instance: {str(e)}")


# gmab/providers/base.py

from abc import ABC, abstractmethod

class ProviderBase(ABC):
    """
    Abstract base class for all providers.
    """

    def __init__(self, provider_cfg):
        """
        Initialize the provider with configuration.
        
        Args:
            provider_cfg (dict): Provider configuration containing credentials, defaults, etc.
                e.g. {
                  "api_key": "...",
                  "default_region": "...",
                  "default_image": "..."
                }
        """
        self.provider_cfg = provider_cfg
        self.provider_name = None  # Will be set by the factory
    
    @staticmethod
    @abstractmethod
    def get_default_config():
        return 

    @staticmethod
    @abstractmethod
    def get_config_prompts(provider_config):
        return []

    @abstractmethod
    def spawn_instance(self, image=None, region=None, ssh_key_path=None, lifetime_minutes=None):
        """
        Create an instance on the provider and return minimal details.
        
        Args:
            image (str, optional): The image to use for the instance
            region (str, optional): The region to create the instance in
            ssh_key_path (str, optional): Path to the SSH public key to use
            lifetime_minutes (int, optional): Lifetime of the instance in minutes
            
        Returns:
            dict: A dictionary with at least these keys:
                {
                    "provider": "provider_name",
                    "instance_id": "provider-specific-id",
                    "label": "instance-name",
                    "ip": "ip-address",
                    "status": "status",
                    "region": "region",
                    "image": "image",
                    "creation_time": unix_timestamp,
                    "lifetime_minutes": minutes
                }
                
        Raises:
            Exception: If the instance creation fails
        """
        pass

    @abstractmethod
    def terminate_instance(self, instance_id):
        """
        Terminate an instance by ID or name.
        
        Args:
            instance_id (str): The ID or name of the instance to terminate
            
        Raises:
            Exception: If the termination fails
        """
        pass

    @abstractmethod
    def list_instances(self):
        """
        List all instances tagged with 'gmab'.
        
        Returns:
            list: A list of dictionaries with the same structure as spawn_instance() returns
            
        Raises:
            Exception: If listing instances fails
        """
        pass

    @abstractmethod
    def list_expired_instances(self):
        """
        List all expired instances.
        
        Returns:
            list: A list of dictionaries with the same structure as list_instances() returns,
                 but only including expired instances
                 
        Raises:
            Exception: If listing expired instances fails
        """
        pass


# gmab/providers/linode.py

import click
import functools
import requests
import random
import string
import time
from pathlib import Path
from gmab.providers.base import ProviderBase

def generate_random_string(length=12):
    """Generate a random string of lowercase letters and digits."""
    return ''.join(random.choices(string.ascii_lowercase + string.digits, k=length))

class LinodeProvider(ProviderBase):
    """
    Provider implementation for Linode.
    """

    @staticmethod
    def get_default_config():
        return {
            "api_key": "",
            "default_region": "nl-ams",
            "default_image": "linode/ubuntu22.04",
            "default_type": "g6-nanode-1",
            "default_root_pass": ""
        }

    @staticmethod
    def get_config_prompts(provider_config):
        config = {}
        config['api_key'] = functools.partial(click.prompt,
            "API Key",
            default=provider_config.get('api_key', ''),
            hide_input=False
        )
        config['default_region'] = functools.partial(click.prompt,
            "Default region",
            default=provider_config.get('default_region', LinodeProvider.get_default_config()['default_region'])
        )
        config['default_image'] = functools.partial(click.prompt,
            "Default image",
            default=provider_config.get('default_image', LinodeProvider.get_default_config()['default_image'])
        )
        config['default_type'] = functools.partial(click.prompt,
            "Default instance type",
            default=provider_config.get('default_type', LinodeProvider.get_default_config()['default_type'])
        )
        config['default_root_pass'] = functools.partial(click.prompt,
            "Default root password",
            default=provider_config.get('default_root_pass', ''),
            hide_input=False
        )
        return config

    def spawn_instance(self, image=None, region=None, ssh_key_path=None, lifetime_minutes=None):
        """
        Create a new Linode instance.
        
        Args:
            image (str, optional): The Linode image to use
            region (str, optional): The region to create the instance in
            ssh_key_path (str, optional): Path to the SSH public key to use
            lifetime_minutes (int, optional): Lifetime of the instance in minutes
            
        Returns:
            dict: Instance information including ID, IP, etc.
            
        Raises:
            ValueError: If API key is missing or other validation fails
            FileNotFoundError: If SSH key file doesn't exist
            Exception: For API errors or other failures
        """
        token = self.provider_cfg.get("api_key")
        if not token:
            raise ValueError("Linode API key not found in config.")

        # Fallback to defaults if arguments are not provided
        linode_type = self.provider_cfg.get("default_type", "g6-nanode-1")
        default_image = self.provider_cfg.get("default_image", "linode/ubuntu22.04")
        default_region = self.provider_cfg.get("default_region", "us-east")

        if image is None:
            print(f"[INFO] No image specified, falling back to default: {default_image}")
            image = default_image
        if region is None:
            print(f"[INFO] No region specified, falling back to default: {default_region}")
            region = default_region
        
        # Set default lifetime if not specified
        if lifetime_minutes is None:
            lifetime_minutes = 60

        # Current timestamp for creation time
        creation_time = int(time.time())

        root_pass = self.provider_cfg.get("default_root_pass", "ChangeMe123!")
        ssh_key_path = ssh_key_path or self.provider_cfg.get("ssh_key_path", "~/.ssh/id_ed25519.pub")

        keyfile = Path(ssh_key_path).expanduser()
        if not keyfile.exists():
            raise FileNotFoundError(f"SSH key not found at {keyfile}")

        with open(keyfile, 'r') as f:
            ssh_key = f.read().strip()

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {token}"
        }

        # Generate a unique Linode name
        random_name = f"gmab-{generate_random_string(12)}"

        data = {
            "type": linode_type,
            "region": region,
            "image": image,
            "authorized_keys": [ssh_key],
            "root_pass": root_pass,
            "label": random_name,
            "tags": [
                "gmab",
                f"gmab-creation-time-{creation_time}",
                f"gmab-lifetime-{lifetime_minutes}"
            ]
        }

        try:
            resp = requests.post(
                "https://api.linode.com/v4/linode/instances",
                headers=headers,
                json=data,
                timeout=30  # Added timeout for better error handling
            )
            
            if resp.status_code not in (200, 202):
                raise Exception(f"Linode creation failed: {resp.text}")

            instance_data = resp.json()
            ip_address = instance_data["ipv4"][0] if instance_data["ipv4"] else "No IP Assigned"
            
            return {
                "provider": self.provider_name,
                "instance_id": str(instance_data["id"]),
                "ip": ip_address,
                "label": random_name,
                "status": instance_data["status"],
                "region": region,
                "image": image,
                "creation_time": creation_time,
                "lifetime_minutes": lifetime_minutes
            }
            
        except requests.exceptions.RequestException as e:
            raise Exception(f"Network error when creating Linode: {str(e)}")
        except Exception as e:
            raise Exception(f"Failed to create Linode instance: {str(e)}")

    def get_instance_id_by_label(self, label):
        """
        Find the Linode instance ID by its label, but only for instances with the 'gmab' tag.
        
        Args:
            label (str): The instance label to search for
            
        Returns:
            str or None: The instance ID if found, None otherwise
        """
        instances = self.list_instances()
        for instance in instances:
            if instance["label"] == label:
                return instance["instance_id"]
        return None

    def terminate_instance(self, instance_identifier):
        """
        Terminate a Linode instance by ID or label.
        
        Args:
            instance_identifier (str): A numeric instance ID or label (e.g., "gmab-3ul7u0p1x9ns")
            
        Raises:
            Exception: If the instance cannot be found or deleted
        """
        token = self.provider_cfg.get("api_key")
        if not token:
            raise ValueError("Linode API key not found in config.")
            
        headers = {"Authorization": f"Bearer {token}"}

        if not instance_identifier.isdigit():
            instance_id = self.get_instance_id_by_label(instance_identifier)
            if instance_id is None:
                raise Exception(f"Instance with label '{instance_identifier}' not found or not tagged with 'gmab'.")
        else:
            instance_id = instance_identifier

        try:
            resp = requests.delete(
                f"https://api.linode.com/v4/linode/instances/{instance_id}",
                headers=headers,
                timeout=30  # Added timeout for better error handling
            )

            if resp.status_code not in (200, 204):
                raise Exception(f"Linode deletion failed: {resp.text}")
                
        except requests.exceptions.RequestException as e:
            raise Exception(f"Network error when terminating Linode: {str(e)}")
        except Exception as e:
            raise Exception(f"Failed to terminate Linode instance: {str(e)}")

    def _get_instance_expiry_info(self, tags):
        """
        Helper method to get expiry information from instance tags.
        
        Args:
            tags (list): List of tags from the Linode instance
            
        Returns:
            tuple: (creation_time, lifetime_minutes, is_expired)
        """
        creation_time = 0
        lifetime_minutes = 60
        
        for tag in tags:
            if tag.startswith("gmab-creation-time-"):
                creation_time = int(tag.split("-")[-1])
            elif tag.startswith("gmab-lifetime-"):
                lifetime_minutes = int(tag.split("-")[-1])
        
        current_time = int(time.time())
        is_expired = (current_time - creation_time) > (lifetime_minutes * 60)
        return creation_time, lifetime_minutes, is_expired

    def list_instances(self):
        """
        Retrieve all active Linode instances that have the 'gmab' tag.
        
        Returns:
            list: List of instance dictionaries
            
        Raises:
            Exception: If listing instances fails
        """
        token = self.provider_cfg.get("api_key")
        if not token:
            raise ValueError("Linode API key not found in config.")
            
        headers = {"Authorization": f"Bearer {token}"}

        try:
            response = requests.get(
                "https://api.linode.com/v4/linode/instances", 
                headers=headers,
                timeout=30  # Added timeout for better error handling
            )
            
            if response.status_code != 200:
                raise Exception(f"Failed to list Linodes: {response.text}")

            instances = response.json()["data"]
            result = []

            for instance in instances:
                if "gmab" in instance.get("tags", []):
                    creation_time, lifetime_minutes, is_expired = self._get_instance_expiry_info(instance.get("tags", []))
                    
                    # Modify status to include expiry information
                    base_status = instance["status"]
                    status = f"{base_status} (expired)" if is_expired else base_status

                    result.append({
                        "provider": self.provider_name,
                        "instance_id": str(instance["id"]),
                        "label": instance["label"],
                        "ip": instance["ipv4"][0] if instance["ipv4"] else "No IP Assigned",
                        "status": status,
                        "region": instance.get("region", "Unknown"),
                        "image": instance.get("image", "Unknown"),
                        "creation_time": creation_time,
                        "lifetime_minutes": lifetime_minutes,
                        "is_expired": is_expired
                    })

            return result
            
        except requests.exceptions.RequestException as e:
            raise Exception(f"Network error when listing Linodes: {str(e)}")
        except Exception as e:
            raise Exception(f"Failed to list Linode instances: {str(e)}")

    def list_expired_instances(self):
        """
        List all expired instances.
        
        Returns:
            list: List of expired instance dictionaries
        """
        return [inst for inst in self.list_instances() if inst['is_expired']]


# gmab/providers/aws.py

import boto3
import click
import functools
import random
import string
import time
from pathlib import Path
from gmab.providers.base import ProviderBase

def generate_random_string(length=12):
    """Generate a random string of lowercase letters and digits."""
    return ''.join(random.choices(string.ascii_lowercase + string.digits, k=length))

class AWSProvider(ProviderBase):
    """
    Provider implementation for AWS EC2.
    """

    def __init__(self, provider_cfg):
        super().__init__(provider_cfg)
        self.session = boto3.Session(
            aws_access_key_id=provider_cfg.get('access_key'),
            aws_secret_access_key=provider_cfg.get('secret_key'),
            region_name=provider_cfg.get('default_region', 'us-east-1')
        )
        self.ec2 = self.session.client('ec2')
        self.ec2_resource = self.session.resource('ec2')

    @staticmethod
    def get_default_config():
        return {
            "access_key": "",
            "secret_key": "",
            "default_region": "eu-west-1",
            "default_image": "ami-0574da719dca65348",
            "default_type": "t3.micro"
        }

    @staticmethod
    def get_config_prompts(provider_config):
        config = {}
        config['access_key'] = functools.partial(click.prompt,
            "Access Key",
            default=provider_config.get('access_key', ''),
            hide_input=False
        )
        config['secret_key'] = functools.partial(click.prompt,
            "Secret Key",
            default=provider_config.get('secret_key', ''),
            hide_input=False
        )
        config['default_region'] = functools.partial(click.prompt,
            "Default region",
            default=provider_config.get('default_region', AWSProvider.get_default_config()['default_region'])
        )
        config['default_image'] = functools.partial(click.prompt,
            "Default image",
            default=provider_config.get('default_image', AWSProvider.get_default_config()['default_image'])
        )
        config['default_type'] = functools.partial(click.prompt,
            "Default instance type",
            default=provider_config.get('default_type', AWSProvider.get_default_config()['default_type'])
        )
        return config

    def get_or_create_vpc(self):
        """Get existing gmab VPC or create a new one."""
        # Check for existing gmab VPC
        vpcs = self.ec2.describe_vpcs(
            Filters=[
                {'Name': 'tag:Name', 'Values': ['gmab-vpc']}
            ]
        )['Vpcs']

        if vpcs:
            return vpcs[0]['VpcId']

        # Create new VPC
        vpc = self.ec2_resource.create_vpc(
            CidrBlock='10.0.0.0/16',
            TagSpecifications=[
                {
                    'ResourceType': 'vpc',
                    'Tags': [
                        {'Key': 'Name', 'Value': 'gmab-vpc'},
                        {'Key': 'gmab', 'Value': 'true'}
                    ]
                }
            ]
        )
        vpc.wait_until_available()

        # Enable DNS hostnames
        self.ec2.modify_vpc_attribute(
            VpcId=vpc.id,
            EnableDnsHostnames={'Value': True}
        )

        # Create and attach internet gateway
        igw = self.ec2_resource.create_internet_gateway(
            TagSpecifications=[
                {
                    'ResourceType': 'internet-gateway',
                    'Tags': [
                        {'Key': 'Name', 'Value': 'gmab-igw'},
                        {'Key': 'gmab', 'Value': 'true'}
                    ]
                }
            ]
        )
        vpc.attach_internet_gateway(InternetGatewayId=igw.id)

        # Create subnet
        subnet = vpc.create_subnet(
            CidrBlock='10.0.1.0/24',
            TagSpecifications=[
                {
                    'ResourceType': 'subnet',
                    'Tags': [
                        {'Key': 'Name', 'Value': 'gmab-subnet'},
                        {'Key': 'gmab', 'Value': 'true'}
                    ]
                }
            ]
        )

        # Create route table and add route to internet
        route_table = vpc.create_route_table(
            TagSpecifications=[
                {
                    'ResourceType': 'route-table',
                    'Tags': [
                        {'Key': 'Name', 'Value': 'gmab-rt'},
                        {'Key': 'gmab', 'Value': 'true'}
                    ]
                }
            ]
        )
        route_table.create_route(
            DestinationCidrBlock='0.0.0.0/0',
            GatewayId=igw.id
        )
        route_table.associate_with_subnet(SubnetId=subnet.id)

        # Enable auto-assign public IP
        self.ec2.modify_subnet_attribute(
            SubnetId=subnet.id,
            MapPublicIpOnLaunch={'Value': True}
        )

        return vpc.id

    def get_or_create_security_group(self, vpc_id):
        """Get existing gmab security group or create a new one."""
        try:
            # Check for existing security group
            security_groups = self.ec2.describe_security_groups(
                Filters=[
                    {'Name': 'group-name', 'Values': ['gmab-sg']},
                    {'Name': 'vpc-id', 'Values': [vpc_id]}
                ]
            )['SecurityGroups']

            if security_groups:
                return security_groups[0]['GroupId']

            # Create new security group
            response = self.ec2.create_security_group(
                GroupName='gmab-sg',
                Description='Security group for GMAB instances',
                VpcId=vpc_id,
                TagSpecifications=[
                    {
                        'ResourceType': 'security-group',
                        'Tags': [
                            {'Key': 'Name', 'Value': 'gmab-sg'},
                            {'Key': 'gmab', 'Value': 'true'}
                        ]
                    }
                ]
            )
            security_group_id = response['GroupId']

            # Add SSH ingress rule
            self.ec2.authorize_security_group_ingress(
                GroupId=security_group_id,
                IpPermissions=[
                    {
                        'IpProtocol': 'tcp',
                        'FromPort': 22,
                        'ToPort': 22,
                        'IpRanges': [{'CidrIp': '0.0.0.0/0'}]
                    }
                ]
            )

            return security_group_id

        except Exception as e:
            raise Exception(f"Failed to setup security group: {str(e)}")

    def get_subnet_id(self, vpc_id):
        """Get the gmab subnet ID for the given VPC."""
        subnets = self.ec2.describe_subnets(
            Filters=[
                {'Name': 'vpc-id', 'Values': [vpc_id]},
                {'Name': 'tag:Name', 'Values': ['gmab-subnet']}
            ]
        )['Subnets']

        if not subnets:
            raise Exception("Could not find gmab subnet")
        
        return subnets[0]['SubnetId']

    def get_instance_id_by_label(self, label):
        """Find instance ID by label, but only for instances with the 'gmab' tag."""
        try:
            response = self.ec2.describe_instances(
                Filters=[
                    {'Name': 'tag:Name', 'Values': [label]},
                    {'Name': 'tag:gmab', 'Values': ['true']},
                    {'Name': 'instance-state-name', 'Values': ['pending', 'running', 'stopping', 'stopped']}
                ]
            )

            for reservation in response['Reservations']:
                for instance in reservation['Instances']:
                    return instance['InstanceId']
            
            return None
        except Exception as e:
            return None

    def _get_instance_expiry_info(self, instance):
        """Helper method to get expiry information from instance tags."""
        tags = {tag['Key']: tag['Value'] for tag in instance.get('Tags', [])}
        creation_time = int(tags.get('gmab-creation-time', '0'))
        lifetime_minutes = int(tags.get('gmab-lifetime', '60'))
        current_time = int(time.time())
        
        is_expired = (current_time - creation_time) > (lifetime_minutes * 60)
        return creation_time, lifetime_minutes, is_expired

    def spawn_instance(self, image=None, region=None, ssh_key_path=None, lifetime_minutes=None):
        # Use provided region or fall back to default
        default_region = self.provider_cfg.get("default_region", "us-east-1")
        chosen_region = region or default_region

        # Use provided image or fall back to default
        default_image = self.provider_cfg.get("default_image", "ami-0574da719dca65348")
        chosen_image = image or default_image

        # Get instance type from config or use default
        instance_type = self.provider_cfg.get("default_type", "t2.micro")

        # Generate a unique name tag
        instance_name = f"gmab-{generate_random_string(12)}"

        # Current timestamp for creation time
        creation_time = int(time.time())
        
        # Default lifetime if not specified
        if lifetime_minutes is None:
            lifetime_minutes = 60

        # Read SSH key
        ssh_key_path = ssh_key_path or self.provider_cfg.get("ssh_key_path", "~/.ssh/id_ed25519.pub")
        keyfile = Path(ssh_key_path).expanduser()
        if not keyfile.exists():
            raise FileNotFoundError(f"SSH key not found at {keyfile}")

        with open(keyfile, 'r') as f:
            ssh_key_content = f.read().strip()

        # Setup networking
        vpc_id = self.get_or_create_vpc()
        security_group_id = self.get_or_create_security_group(vpc_id)
        subnet_id = self.get_subnet_id(vpc_id)

        # Import SSH key to AWS
        key_name = f"gmab-key-{generate_random_string(8)}"
        try:
            self.ec2.import_key_pair(
                KeyName=key_name,
                PublicKeyMaterial=ssh_key_content.encode()
            )
        except Exception as e:
            raise Exception(f"Failed to import SSH key to AWS: {str(e)}")

        # Launch instance with standardized tags
        try:
            response = self.ec2.run_instances(
                ImageId=chosen_image,
                InstanceType=instance_type,
                MinCount=1,
                MaxCount=1,
                KeyName=key_name,
                SecurityGroupIds=[security_group_id],
                SubnetId=subnet_id,
                TagSpecifications=[
                    {
                        'ResourceType': 'instance',
                        'Tags': [
                            {'Key': 'Name', 'Value': instance_name},
                            {'Key': 'gmab', 'Value': 'true'},
                            {'Key': 'gmab-creation-time', 'Value': str(creation_time)},
                            {'Key': 'gmab-lifetime', 'Value': str(lifetime_minutes)}
                        ]
                    }
                ]
            )
        except Exception as e:
            # Clean up the key pair if instance launch fails
            try:
                self.ec2.delete_key_pair(KeyName=key_name)
            except:
                pass
            raise Exception(f"Failed to launch AWS instance: {str(e)}")

        instance = response['Instances'][0]
        instance_id = instance['InstanceId']

        try:
            # Wait for instance to be running and get its public IP
            waiter = self.ec2.get_waiter('instance_running')
            waiter.wait(InstanceIds=[instance_id])

            # Get instance details
            instance_info = self.ec2.describe_instances(InstanceIds=[instance_id])['Reservations'][0]['Instances'][0]
            ip_address = instance_info.get('PublicIpAddress', 'No IP Assigned')

            return {
                "provider": self.provider_name,
                "instance_id": instance_id,
                "label": instance_name,
                "ip": ip_address,
                "status": instance_info['State']['Name'],
                "region": chosen_region,
                "image": chosen_image,
                "creation_time": creation_time,
                "lifetime_minutes": lifetime_minutes
            }
        except Exception as e:
            # Clean up on error
            try:
                self.ec2.terminate_instances(InstanceIds=[instance_id])
                self.ec2.delete_key_pair(KeyName=key_name)
            except:
                pass
            raise Exception(f"Failed to get instance details: {str(e)}")

    def terminate_instance(self, instance_identifier):
        """
        Terminate an EC2 instance by ID or label.
        Args:
            instance_identifier: Can be either an instance ID (i-xxxxx) or a label (gmab-xxxxx)
        """
        try:
            # If it's not a typical AWS instance ID format, try to find by label
            if not instance_identifier.startswith('i-'):
                instance_id = self.get_instance_id_by_label(instance_identifier)
                if instance_id is None:
                    raise Exception(f"No instance found with label '{instance_identifier}'")
            else:
                instance_id = instance_identifier

            # Get the instance tags to find our key pair name if it exists
            response = self.ec2.describe_instances(InstanceIds=[instance_id])
            instance = response['Reservations'][0]['Instances'][0]
            
            # Find any gmab key pairs associated with this instance
            key_name = instance.get('KeyName')
            if key_name and key_name.startswith('gmab-key-'):
                try:
                    self.ec2.delete_key_pair(KeyName=key_name)
                except:
                    pass  # Best effort cleanup

            self.ec2.terminate_instances(InstanceIds=[instance_id])
        except Exception as e:
            raise Exception(f"Failed to terminate AWS instance: {str(e)}")

    def list_instances(self):
        """List all EC2 instances tagged with 'gmab'."""
        try:
            response = self.ec2.describe_instances(
                Filters=[
                    {'Name': 'tag:gmab', 'Values': ['true']},
                    {'Name': 'instance-state-name', 'Values': ['pending', 'running', 'stopping', 'stopped']}
                ]
            )

            instances = []
            for reservation in response['Reservations']:
                for instance in reservation['Instances']:
                    name_tag = next((tag['Value'] for tag in instance.get('Tags', []) if tag['Key'] == 'Name'), 'Unknown')
                    creation_time, lifetime_minutes, is_expired = self._get_instance_expiry_info(instance)
                    
                    # Modify status to include expiry information
                    base_status = instance['State']['Name']
                    status = f"{base_status} (expired)" if is_expired else base_status

                    instances.append({
                        "provider": self.provider_name,
                        "instance_id": instance['InstanceId'],
                        "label": name_tag,
                        "ip": instance.get('PublicIpAddress', 'No IP Assigned'),
                        "status": status,
                        "region": instance['Placement']['AvailabilityZone'][:-1],
                        "image": instance['ImageId'],
                        "creation_time": creation_time,
                        "lifetime_minutes": lifetime_minutes,
                        "is_expired": is_expired
                    })

            return instances

        except Exception as e:
            raise Exception(f"Failed to list AWS instances: {str(e)}")

    def list_expired_instances(self):
        """List all expired instances."""
        return [inst for inst in self.list_instances() if inst['is_expired']]


# gmab/utils/paths.py

import os
import sys
from pathlib import Path

def get_config_dir():
    """
    Get the configuration directory following XDG Base Directory Specification.
    Falls back to platform-specific user config locations.
    
    Returns:
        Path: Path object representing the config directory
    """
    # First check if GMAB_CONFIG_DIR environment variable is set
    if 'GMAB_CONFIG_DIR' in os.environ:
        return Path(os.environ['GMAB_CONFIG_DIR'])

    # On Unix-like systems, follow XDG specification
    if sys.platform.startswith('linux') or sys.platform.startswith('darwin'):
        xdg_config_home = os.environ.get('XDG_CONFIG_HOME', os.path.expanduser('~/.config'))
        return Path(xdg_config_home) / 'gmab'

    # On Windows, use %APPDATA%
    elif sys.platform == 'win32':
        return Path(os.environ['APPDATA']) / 'gmab'

    # Fallback to user's home directory
    return Path.home() / '.gmab'

def get_config_file_path(filename):
    """
    Get the full path for a config file.
    
    Args:
        filename (str): The filename
        
    Returns:
        Path: Path object representing the config file path
    """
    config_dir = get_config_dir()
    return config_dir / filename

def ensure_config_dir_exists():
    """
    Create the config directory if it doesn't exist.
    
    Returns:
        Path: Path object representing the config directory
    """
    config_dir = get_config_dir()
    config_dir.mkdir(parents=True, exist_ok=True)
    return config_dir